package de.andrena.objektforum.timeconcepts;

import static java.time.Month.FEBRUARY;
import static java.time.Month.JANUARY;
import static java.time.Month.JUNE;
import static java.time.Month.MARCH;
import static java.time.Month.MAY;
import static java.time.temporal.ChronoUnit.DAYS;
import static java.time.temporal.ChronoUnit.MINUTES;
import static java.time.temporal.ChronoUnit.MONTHS;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import java.text.SimpleDateFormat;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.Period;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.TimeZone;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

public class AllExamples {

    @Test
    void localDateTime() {
        LocalDateTime dateTime = LocalDateTime.of(2020, JUNE, 15, 18, 31, 42);

        System.out.println(dateTime);
        System.out.println(dateTime.plus(31, DAYS));
        System.out.println(dateTime.plus(1, MONTHS));
        System.out.println();
        System.out.println(formatLocalDateTime(dateTime));
    }

    @Test
    void instant() {
        Instant now = Instant.now();

        System.out.println(now); // ISO String in UTC
        System.out.println(now.plus(31, DAYS));
        assertThatThrownBy(() -> now.plus(1, MONTHS)).hasMessageContaining("Unsupported unit: Months");

        System.out.println();
        System.out.println(formatInstant(now));
    }

    @Test
    void instantAndEpochMilli() {
        long currentTimeMillis = System.currentTimeMillis();
        Instant instantFromMillis = Instant.ofEpochMilli(currentTimeMillis);
        Instant instantNow = Instant.now();

        System.out.println(currentTimeMillis);
        System.out.println(instantFromMillis.toEpochMilli());
        System.out.println(instantNow.toEpochMilli());
    }

    @Test
    void instantToLocalDateTime() {
        Instant now = Instant.now();
        System.out.println(formatInstant(now));
        System.out.println();

        LocalDateTime nowInUTC = LocalDateTime.ofInstant(now, ZoneId.of("UTC"));
        LocalDateTime nowInNYC = LocalDateTime.ofInstant(now, ZoneId.of("America/New_York"));
        LocalDateTime nowHere = LocalDateTime.ofInstant(now, ZoneId.systemDefault());
        LocalDateTime nowInSingapore = LocalDateTime.ofInstant(now, ZoneId.of("Asia/Singapore"));

        System.out.println(formatLocalDateTime(nowInUTC) + " (now in UTC)");
        System.out.println(formatLocalDateTime(nowInNYC) + " (now in New York)");
        System.out.println(formatLocalDateTime(nowHere) + " (now here)");
        System.out.println(formatLocalDateTime(nowInSingapore) + " (now in Singapore)");
    }

    @Test
    void currentSystemDateTime() {
        LocalDateTime currentSystemDateTime = LocalDateTime.ofInstant(Instant.now(), ZoneId.systemDefault());

        System.out.println(currentSystemDateTime);
        System.out.println(LocalDateTime.now()); // just a shortcut!
    }

    @Test
    void localDateTimeToInstant() {
        LocalDateTime newYear = LocalDateTime.of(2020, JANUARY, 1, 0, 0, 0);
        System.out.println(formatLocalDateTime(newYear));
        System.out.println();

        Instant newYearInLondon = newYear.atZone(ZoneId.of("Europe/London")).toInstant();
        Instant newYearInNYC = newYear.atZone(ZoneId.of("America/New_York")).toInstant();
        Instant newYearInSystemDefault = newYear.atZone(ZoneId.systemDefault()).toInstant();
        Instant newYearInSingapore = newYear.atZone(ZoneId.of("Asia/Singapore")).toInstant();

        System.out.println(formatInstant(newYearInLondon) + " (London's new year)");
        System.out.println(formatInstant(newYearInNYC) + " (New York's new year)");
        System.out.println(formatInstant(newYearInSystemDefault) + " (our new year)");
        System.out.println(formatInstant(newYearInSingapore) + " (Singapore's new year)");
    }

    @Test
    void zonedDateTime() {
        ZonedDateTime zonedDateTime = LocalDateTime
            .of(2020, JUNE, 4, 12, 42, 42)
            .atZone(ZoneId.of("Europe/Berlin"));

        System.out.println(zonedDateTime);
        System.out.println();
        System.out.println(zonedDateTime.toLocalDateTime() + " (local date time)");
        System.out.println(zonedDateTime.toInstant() + " (instant)");
    }

    @Nested
    class offsetDateTimeIgnoresWinterAndSummerTime {

        @Test
        void zoneAndOffsetMatchInWinter() {
            LocalDateTime duringWinterTime = LocalDateTime.of(2020, FEBRUARY, 4, 12, 42, 42);
            ZonedDateTime usingZoneId = duringWinterTime.atZone(ZoneId.of("Europe/Berlin"));
            OffsetDateTime usingOffset = duringWinterTime.atOffset(ZoneOffset.ofHours(1));

            System.out.println(usingZoneId);
            System.out.println(usingOffset);
            assertThat(usingZoneId.toInstant()).isEqualTo(usingOffset.toInstant());
        }

        @Test
        void zoneAndOffsetDifferInSummer() {
            LocalDateTime duringSummerTime = LocalDateTime.of(2020, JUNE, 4, 12, 42, 42);
            ZonedDateTime usingZoneId = duringSummerTime.atZone(ZoneId.of("Europe/Berlin"));
            OffsetDateTime usingOffset = duringSummerTime.atOffset(ZoneOffset.ofHours(1));

            System.out.println(usingZoneId);
            System.out.println(usingOffset);
            assertThat(usingZoneId.toInstant()).isNotEqualTo(usingOffset.toInstant());
        }
    }

    @Test
    void switchToSummerTime() {
        ZonedDateTime oneMinuteBeforeSwitchInGermany = LocalDateTime.of(2020, 3, 29, 1, 59, 0, 0)
            .atZone(ZoneId.of("Europe/Berlin"));
        LocalDateTime pureDateTime = oneMinuteBeforeSwitchInGermany.toLocalDateTime();
        Instant asInstant = oneMinuteBeforeSwitchInGermany.toInstant();

        System.out.println(formatLocalDateTime(pureDateTime));
        System.out.println(formatLocalDateTime(pureDateTime.plus(2, MINUTES)));

        System.out.println();
        System.out.println(formatInstant(asInstant));
        System.out.println(formatInstant(asInstant.plus(2, MINUTES)));

        System.out.println();
        System.out.println(oneMinuteBeforeSwitchInGermany);
        System.out.println(oneMinuteBeforeSwitchInGermany.plus(2, MINUTES));
    }

    @Nested
    class formatting {

        @Test
        void formattingInstant_requiresZoneId() {
            DateTimeFormatter formatter = DateTimeFormatter
                .ofPattern("dd.MM.yyyy HH:mm")
                .withZone(ZoneId.of("Asia/Singapore"));

            Instant now = Instant.now();
            System.out.println(formatter.format(now));
        }

        @Test
        void formattingLocalDateTime_noNeedForZoneId() {
            DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy HH:mm");

            LocalDateTime localDateTime = LocalDateTime.of(2020, 5, 12, 16, 30);
            System.out.println(formatter.format(localDateTime));
        }

        @Test
        void formattingZonedDateTime() {
            DateTimeFormatter formatterWithoutZoneId = DateTimeFormatter
                .ofPattern("dd.MM.yyyy HH:mm");
            DateTimeFormatter formatterForSingaporeTime = DateTimeFormatter
                .ofPattern("dd.MM.yyyy HH:mm")
                .withZone(ZoneId.of("Asia/Singapore"));

            ZonedDateTime zonedDateTime = LocalDateTime.of(2020, 5, 12, 16, 30).atZone(ZoneId.of("Europe/Berlin"));
            System.out.println(formatterWithoutZoneId.format(zonedDateTime));
            System.out.println(formatterForSingaporeTime.format(zonedDateTime));
        }
    }

    @Test
    void parsing() {
        DateTimeFormatter formatter = DateTimeFormatter
            .ofPattern("dd.MM.yyyy HH:mm")
            .withZone(ZoneId.of("Asia/Singapore"));

        Instant parsed = Instant.from(formatter.parse("01.01.2020 00:00"));
        System.out.println(formatInstant(parsed));
    }

    @Test
    void duration() {
        Instant end = Instant.now();
        Instant start = end.minus(80, DAYS);

        long days = Duration.between(start, end).toDays();
        assertThat(days).isEqualTo(80);
    }

    @Test
    void period() {
        LocalDate start = LocalDate.of(2020, MARCH, 12);
        LocalDate end = LocalDate.of(2020, MAY, 12);

        long months = Period.between(start, end).getMonths();
        assertThat(months).isEqualTo(2);
    }

    @Test
    void changeSystemTimeZone() {
        TimeZone timeZoneBefore = TimeZone.getDefault();

        Instant nowHere = Instant.now();
        LocalDateTime currentDateTimeHere = LocalDateTime.now();

        TimeZone.setDefault(TimeZone.getTimeZone("Asia/Singapore"));

        Instant nowInSingapore = Instant.now();
        LocalDateTime currentDateTimeInSingapore = LocalDateTime.now();

        System.out.println(formatInstant(nowHere));
        System.out.println(formatInstant(nowInSingapore));
        System.out.println(formatLocalDateTime(currentDateTimeHere));
        System.out.println(formatLocalDateTime(currentDateTimeInSingapore));

        TimeZone.setDefault(timeZoneBefore);
    }

    @Nested
    class javaUtilDate {

        @Test
        void now() {
            System.out.println(new Date().getTime());
            System.out.println(new Date().toInstant().toEpochMilli());
            System.out.println(Instant.now().toEpochMilli());
        }

        @Test
        void implicitSystemTimeZone() {
            Date dateFromString = new Date("Tue, 15 June 2020 16:05:00 UTC");
            Date dateFromCoordinates = new Date(120, 5, 15, 18, 5, 0);
            ZonedDateTime zonedDateTime = LocalDateTime.of(2020, 6, 15, 18, 5, 0).atZone(ZoneId.systemDefault());

            System.out.println(dateFromString);
            System.out.println(dateFromCoordinates);
            System.out.println((dateFromCoordinates.getYear() + 1900) + " " + (dateFromCoordinates.getMonth() + 1) + " " + (dateFromCoordinates.getDate()));
            System.out.println(zonedDateTime.getYear() + " " + zonedDateTime.getMonthValue() + " " + zonedDateTime.getDayOfMonth());
        }

        @Test
        void formattingDateTime() {
            SimpleDateFormat simpleDateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm");
            simpleDateFormat.setTimeZone(TimeZone.getTimeZone("Asia/Singapore"));
            Date date = new Date(120, 5, 15, 18, 5, 0);
            System.out.println(simpleDateFormat.format(date) + " (with Date and SimpleDateFormat)");

            DateTimeFormatter dateTimeFormat = DateTimeFormatter
                .ofPattern("dd.MM.yyyy HH:mm")
                .withZone(ZoneId.of("Asia/Singapore"));
            ZonedDateTime dateTime = LocalDateTime.of(2020, 6, 15, 18, 5, 0).atZone(ZoneId.systemDefault());

            System.out.println(dateTimeFormat.format(dateTime) + " (with ZonedDateTime and DateTimeFormatter)");
        }

        @Test
        void formattingGlobalTimestamp() {
            SimpleDateFormat simpleDateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm");
            simpleDateFormat.setTimeZone(TimeZone.getTimeZone("Asia/Singapore"));
            Date date = new Date(12345678900L);

            DateTimeFormatter dateTimeFormat = DateTimeFormatter
                .ofPattern("dd.MM.yyyy HH:mm")
                .withZone(ZoneId.of("Asia/Singapore"));
            Instant instant = Instant.ofEpochMilli(12345678900L);

            System.out.println(simpleDateFormat.format(date) + " (with Date and SimpleDateFormat)");
            System.out.println(dateTimeFormat.format(instant) + " (with Instant and DateTimeFormatter)");
        }
    }

    private static String formatLocalDateTime(LocalDateTime localDateTime) {
        return DateTimeFormatter
            .ofPattern("yyyy-MM-dd HH:mm:ss")
            .format(localDateTime);
    }

    public static String formatInstant(Instant instant) {
        return DateTimeFormatter
            .ofPattern("yyyy-MM-dd HH:mm:ss")
            .withZone(ZoneId.of("UTC"))
            .format(instant)
            + " UTC [instant, " + instant.toEpochMilli() + "]";
    }
}
